import os


class Panel(object):

    def __init__(self, filepath, type="sequence"):
        self.filepath = filepath
        self.name = os.path.basename(filepath).split('.')[0]
        self.type = type

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name



from math import exp
from math import factorial
from math import log


def percent_coverage_from_expected_coverage(coverage):
    # With low coverage we expect a lower percent of the sequence to be
    # coverage.
    return 1 - exp(-coverage)


def log_poisson_prob(lam, k):
    return -lam + k * log(lam) - log_factorial(k)


def log_factorial(n):
    assert n >= 0
    out = 0
    for i in range(int(n)):
        out += log(i + 1)
    return out


def log_lik_depth(depth, expected_depth):
    if expected_depth <= 0:
        raise ValueError("Expected depth must be greater than 0")
    if depth < 0:
        raise ValueError("Depth must not be negative")
    return log_poisson_prob(lam=expected_depth, k=depth)


def log_lik_R_S_coverage(observed_alternate_depth,
                         observed_reference_depth,
                         expected_alternate_depth,
                         expected_reference_depth):
    lne = log_poisson_prob(
        lam=expected_alternate_depth,
        k=observed_alternate_depth)
    le = log_poisson_prob(
        lam=expected_reference_depth,
        k=observed_reference_depth)
    return lne + le



class ProbeCoverage(object):

    "Summary of kmer coverage of sequence. e.g output of color covearges"

    def __init__(self, percent_coverage, median_depth, min_depth):
        self.percent_coverage = percent_coverage
        self.median_depth = median_depth
        self.min_depth = min_depth

    @property
    def coverage_dict(self):
        return {"percent_coverage": round(self.percent_coverage, 2),
                "median_depth": round(self.median_depth, 2),
                "min_non_zero_depth": round(self.min_depth, 2),
                }



import datetime
import json


class VariantProbeCoverage(object):

    def __init__(self, reference_coverage,
                 alternate_coverages,
                 var_name=None,
                 params={}):
        self.reference_coverage = reference_coverage
        self.alternate_coverages = alternate_coverages
        self.var_name = var_name
        self.params = params
        self.best_alternate_coverage = self._choose_best_alternate_coverage()

    def _choose_best_alternate_coverage(self):
        self.alternate_coverages.sort(
            key=lambda x: x.percent_coverage,
            reverse=True)
        current_best = self.alternate_coverages[0]
        for probe_coverage in self.alternate_coverages[1:]:
            if probe_coverage.percent_coverage < current_best.percent_coverage:
                return current_best
            else:
                if probe_coverage.min_depth > current_best.min_depth:
                    current_best = probe_coverage
                elif probe_coverage.min_depth <= current_best.min_depth:
                    if probe_coverage.median_depth > current_best.median_depth:
                        current_best = probe_coverage
        return current_best

    @property
    def coverage_dict(self):
        return {"reference": self.reference_coverage.coverage_dict,
                "alternate": self.best_alternate_coverage.coverage_dict
                }

    @property
    def reference_percent_coverage(self):
        return self.reference_coverage.percent_coverage

    @property
    def reference_median_depth(self):
        return self.reference_coverage.median_depth

    @property
    def reference_min_depth(self):
        return self.reference_coverage.min_depth

    @property
    def alternate_percent_coverage(self):
        return self.best_alternate_coverage.percent_coverage

    @alternate_percent_coverage.setter
    def alternate_percent_coverage(self, value):
        self.best_alternate_coverage.percent_coverage = value

    @property
    def alternate_median_depth(self):
        return self.best_alternate_coverage.median_depth

    @property
    def alternate_min_depth(self):
        return self.best_alternate_coverage.min_depth



from mykatlas.stats.stats import percent_coverage_from_expected_coverage
from mykatlas.stats.stats import log_factorial
from mykatlas.stats.stats import log_lik_depth
from mykatlas.stats.stats import log_lik_R_S_coverage



#! /usr/bin/env python
import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/..")
import datetime
import logging
logger = logging.getLogger(__name__)


from mongoengine import connect

from ga4ghmongo.schema import Variant
from mykatlas.panelgeneration import AlleleGenerator
from mykatlas.panelgeneration import make_variant_probe


def run(parser, args):
    DB = connect('atlas-%s' % (args.db_name))
    if args.verbose:
        logger.setLevel(level=logging.DEBUG)
    else:
        logger.setLevel(level=logging.ERROR)
    al = AlleleGenerator(
        reference_filepath=args.reference_filepath,
        kmer=args.kmer)
    for variant in Variant.snps():
        variant_panel = make_variant_probe(al, variant, args.kmer, DB=DB)
        sys.stdout.write(
            ">ref-%s?num_alts=%i&ref=%s\n" %
            (variant_panel.variant.var_hash, len(
                variant_panel.alts), variant_panel.variant.reference.id))
        sys.stdout.write("%s\n" % variant_panel.ref)
        for a in variant_panel.alts:
            sys.stdout.write(">alt-%s\n" % variant_panel.variant.var_hash)
            sys.stdout.write("%s\n" % a)



from mykatlas.typing.models.base import ProbeCoverage
from mykatlas.typing.models.presence import SequenceProbeCoverage
from mykatlas.typing.models.panel import Panel
from mykatlas.typing.models.variant import VariantProbeCoverage
from mykatlas.typing.typer.presence import PresenceTyper
from mykatlas.typing.typer.variant import VariantTyper
from mykatlas.typing.typer.genotyper import Genotyper
from mykatlas.typing.typer.genotyper import CoverageParser



#! /usr/bin/env python
from __future__ import print_function
import sys
import os


import os
import argparse
import sys
sys.path.append(
    os.path.realpath(
        os.path.join(
            os.path.dirname(__file__),
            "..")))
from mykatlas.version import __version__
import logging

from base import ArgumentParserWithDefaults
from base import DEFAULT_DB_NAME
from base import sequence_parser_mixin
DEFAULT_KMER_SIZE = os.environ.get("KMER_SIZE", 31)


def run_subtool(parser, args):
    if args.command == 'add':
        from mykatlas.cmds.add import run
    elif args.command == "dump-probes":
        from mykatlas.cmds.dump import run
    elif args.command == "make-probes":
        from mykatlas.cmds.makeprobes import run
    elif args.command == "genotype":
        from mykatlas.cmds.genotype import run
    # run the chosen submodule.
    run(parser, args)


def main():
    #########################################
    # create the top-level parser
    #########################################
    parser = argparse.ArgumentParser(
        prog='atlas',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--version", help="atlas version",
                        action="version",
                        version="%(prog)s " + str(__version__))
    subparsers = parser.add_subparsers(
        title='[sub-commands]',
        dest='command',
        parser_class=ArgumentParserWithDefaults)

    db_parser_mixin = argparse.ArgumentParser(add_help=False)
    db_parser_mixin.add_argument(
        '--db_name',
        metavar='db_name',
        type=str,
        help='db_name',
        default=None)

    ##########
    # Add
    ##########
    parser_add = subparsers.add_parser(
        'add',
        help='adds a set of variants to the atlas',
        parents=[db_parser_mixin])
    parser_add.add_argument('vcf', type=str, help='a vcf file')
    parser_add.add_argument('reference_set', type=str, help='reference set')
    parser_add.add_argument(
        '-m',
        '--method',
        type=str,
        help='variant caller method (e.g. CORTEX)',
        default="NotSpecified")
    parser_add.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='Force recreate VariantSet')
    parser_add.set_defaults(func=run_subtool)

    # ##########
    # # Dump panel
    # ##########
    parser_dump = subparsers.add_parser(
        'dump-probes',
        help='dump a probe set of variant alleles',
        parents=[db_parser_mixin])
    parser_dump.add_argument(
        'reference_filepath',
        metavar='reference_filepath',
        type=str,
        help='reference_filepath')
    parser_dump.add_argument(
        '--kmer',
        metavar='kmer',
        type=int,
        help='kmer length',
        default=31)
    parser_dump.add_argument('--force', default=False, action="store_true")
    parser_dump.add_argument(
        '-v',
        '--verbose',
        default=False,
        action="store_true")
    parser_dump.set_defaults(func=run_subtool)

    ##################
    ### Make Probes ##
    ##################

    parser_make_probes = subparsers.add_parser(
        'make-probes', help='make probes from a list of variants',
        parents=[db_parser_mixin])
    parser_make_probes.add_argument(
        'reference_filepath',
        metavar='reference_filepath',
        type=str,
        help='reference_filepath')
    parser_make_probes.add_argument(
        '-v',
        '--variant',
        type=str,
        action='append',
        help='Variant in DNA positions e.g. A1234T',
        default=[])
    parser_make_probes.add_argument(
        '-f',
        '--file',
        type=str,
        help='File containing variants as rows A1234T')
    parser_make_probes.add_argument(
        '-g',
        '--genbank',
        type=str,
        help='Genbank file containing genes as features')
    parser_make_probes.add_argument(
        '-k',
        '--kmer',
        type=int,
        help='kmer length',
        default=31)
    parser_make_probes.add_argument(
        '--no-backgrounds',
        help='Build probe set against reference only ignoring nearby variants',
        default=False,
        action="store_true")
    parser_make_probes.set_defaults(func=run_subtool)

    # ##########
    # # Genotype
    # ##########
    parser_geno = subparsers.add_parser(
        'genotype',
        parents=[sequence_parser_mixin],
        help='genotype a sample using a probe set')
    parser_geno.add_argument(
        'probe_sets',
        metavar='parser_geno',
        type=str,
        nargs='+',
        help='probe-set')
    parser_geno.add_argument(
        '--expected_depth',
        metavar='expected depth',
        type=int,
        help='expected depth',
        default=None)
    parser_geno.add_argument(
        '-f',
        '--force',
        help='Force rebuilding of binaries',
        default=False,
        action="store_true")
    parser_geno.add_argument(
        '-t',
        '--threads',
        type=int,
        help='threads',
        default=2)
    parser_geno.add_argument(
        '--ignore_filtered',
        help="don't include filtered genotypes",
        default=False)
    parser_geno.set_defaults(func=run_subtool)

    args = parser.parse_args()
    args.func(parser, args)


if __name__ == "__main__":
    main()



MIN_CONF = -99999999
MIN_LLK = -99999999


class Typer(object):

    def __init__(
            self,
            expected_depths,
            contamination_depths=[],
            error_rate=0.05,
            ignore_filtered=False):
        self.expected_depths = expected_depths
        self.contamination_depths = contamination_depths
        self.error_rate = error_rate
        self.ignore_filtered = ignore_filtered

    def type(self, l):
        raise NotImplemented("Implemented in sub class")

    def likelihoods_to_genotype(self, likelihoods, min_conf=MIN_CONF):
        ml = max(likelihoods)
        i = likelihoods.index(ml)
        if i == 0:
            if ml <= min_conf:
                gt = "-/-"
            else:
                gt = "0/0"
        elif i == 1:
            gt = "0/1"
        elif i == 2:
            gt = "1/1"
        return gt

    def has_contamination(self):
        return self.contamination_depths or len(self.expected_depths) > 1



from setuptools import setup
from mykatlas.version import __version__
__version__ = __version__[1:].split("-")[0]
setup(
    name='mykatlas',
    version=__version__,
    packages=[
        'mykatlas',
        'mykatlas.cmds',
        'mykatlas.typing',
        'mykatlas.typing.typer',
        'mykatlas.typing.models',
        'mykatlas.stats',
        'mykatlas.cortex'],
    license='MIT',
    url='http://github.com/phelimb/atlas',
    description='.',
    author='Phelim Bradley',
    author_email='wave@phel.im',
    install_requires=[
            'future',
            'Biopython',
            'mongoengine',
            'pyvcf',
            'ga4ghmongo'],
    entry_points={
        'console_scripts': [
            'atlas = atlas.atlas_main:main',
        ]})



from mykatlas.typing.typer.base import Typer
from mykatlas.stats import log_lik_R_S_coverage
from mykatlas.typing.typer.base import MIN_LLK

from ga4ghmongo.schema import VariantCall

DEFAULT_ERROR_RATE = 0.05
DEFAULT_MINOR_FREQ = 0.2
from mykatlas.stats import percent_coverage_from_expected_coverage


class VariantTyper(Typer):

    def __init__(self, expected_depths, contamination_depths=[],
                 error_rate=DEFAULT_ERROR_RATE,
                 minor_freq=DEFAULT_MINOR_FREQ,
                 ignore_filtered=False):
        super(
            VariantTyper,
            self).__init__(
            expected_depths,
            contamination_depths,
            error_rate,
            ignore_filtered=ignore_filtered)
        self.method = "MAP"
        self.error_rate = error_rate
        self.minor_freq = minor_freq
        self.ignore_filtered = ignore_filtered

        if len(expected_depths) > 1:
            raise NotImplementedError("Mixed samples not handled yet")

    def type(self, variant_probe_coverages, variant=None):
        """
            Takes a list of VariantProbeCoverages and returns a Call for the Variant.
            Note, in the simplest case the list will be of length one. However, we may be typing the
            Variant on multiple backgrouds leading to multiple VariantProbes for a single Variant.

        """
        if not isinstance(variant_probe_coverages, list):
            variant_probe_coverages = [variant_probe_coverages]
        calls = []
        for variant_probe_coverage in variant_probe_coverages:
            calls.append(
                self._type_variant_probe_coverages(
                    variant_probe_coverage, variant))
        hom_alt_calls = [c for c in calls if sum(c.genotype) > 1]
        het_calls = [c for c in calls if sum(c.genotype) == 1]
        if hom_alt_calls:
            hom_alt_calls.sort(key=lambda x: x.genotype_conf, reverse=True)
            return hom_alt_calls[0]
        elif het_calls:
            het_calls.sort(key=lambda x: x.genotype_conf, reverse=True)
            return het_calls[0]
        else:
            calls.sort(key=lambda x: x.genotype_conf, reverse=True)
            return calls[0]

    def _type_variant_probe_coverages(
            self, variant_probe_coverage, variant=None):
        variant_probe_coverage = self._check_min_coverage(
            variant_probe_coverage)
        hom_ref_likelihood = self._hom_ref_lik(variant_probe_coverage)
        hom_alt_likelihood = self._hom_alt_lik(variant_probe_coverage)
        if not self.has_contamination():
            het_likelihood = self._het_lik(variant_probe_coverage)
        else:
            het_likelihood = MIN_LLK
        likelihoods = [hom_ref_likelihood, het_likelihood, hom_alt_likelihood]
        gt = self.likelihoods_to_genotype(
            likelihoods
        )
        info = {"coverage": variant_probe_coverage.coverage_dict,
                "expected_depths": self.expected_depths,
                "contamination_depths": self.contamination_depths,
                "filter": "PASS"}
        if gt == "-/-" and not self.ignore_filtered:
            if variant_probe_coverage.alternate_percent_coverage > variant_probe_coverage.reference_percent_coverage:
                gt = "1/1"
                info["filter"] = "MISSING_WT"
            else:
                gt = "0/0"
                info["filter"] = "MISSING_WT"
        return VariantCall.create(
            variant=variant,
            genotype=gt,
            genotype_likelihoods=likelihoods,
            info=info)

    def _check_min_coverage(self, variant_probe_coverage):
        if variant_probe_coverage.alternate_min_depth < 0.1 * \
                max(self.expected_depths):
            variant_probe_coverage.alternate_percent_coverage = variant_probe_coverage.alternate_percent_coverage * 0.9
        return variant_probe_coverage

    def _hom_ref_lik(self, variant):
        if variant.reference_percent_coverage < 100 * \
                percent_coverage_from_expected_coverage(max(self.expected_depths)):
            return MIN_LLK
        else:
            hom_ref_likes = []
            # Either alt+cov or alt_covg + contam_covg
            for expected_depth in self.expected_depths:
                hom_ref_likes.append(
                    log_lik_R_S_coverage(
                        variant.reference_median_depth,
                        variant.alternate_median_depth,
                        expected_depth,
                        expected_depth *
                        self.error_rate /
                        3))
                for contamination in self.contamination_depths:
                    hom_ref_likes.append(
                        log_lik_R_S_coverage(
                            variant.reference_median_depth,
                            variant.alternate_median_depth,
                            expected_depth + contamination,
                            (expected_depth + contamination) * self.error_rate / 3))
            return max(hom_ref_likes)

    def _hom_alt_lik(self, variant):
        if variant.alternate_percent_coverage < 100 * \
                percent_coverage_from_expected_coverage(max(self.expected_depths)):
            return MIN_LLK
        else:
            hom_alt_liks = []
            # Either alt+cov or alt_covg + contam_covg
            for expected_depth in self.expected_depths:
                hom_alt_liks.append(
                    log_lik_R_S_coverage(
                        variant.alternate_median_depth,
                        variant.reference_median_depth,
                        expected_depth,
                        expected_depth *
                        self.error_rate /
                        3))
                for contamination in self.contamination_depths:
                    hom_alt_liks.append(
                        log_lik_R_S_coverage(
                            variant.alternate_median_depth,
                            variant.reference_median_depth,
                            expected_depth + contamination,
                            (expected_depth + contamination) * self.error_rate / 3))
            return max(hom_alt_liks)

    def _het_lik(self, variant):
        if variant.alternate_percent_coverage < 100 or variant.reference_percent_coverage < 100:
            return MIN_LLK
        else:
            het_liks = []
            for expected_depth in self.expected_depths:
                het_liks.append(
                    log_lik_R_S_coverage(
                        variant.alternate_median_depth,
                        variant.reference_median_depth,
                        expected_depth * self.minor_freq,
                        expected_depth * (
                            1 - self.minor_freq)))
            return max(het_liks)
